import 'dart:async';
import 'dart:io' show Platform;
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:media_projection_creator/media_projection_creator.dart';
import 'package:vdotok_stream/src/signaling/web_socket_connection.dart';
import 'package:vdotok_stream/src/utils/config.dart';
import 'dart:convert';
import 'package:crypto/crypto.dart';
import 'package:replay_kit_launcher/replay_kit_launcher.dart';
import 'package:http/http.dart' as http;

typedef void OnConnectCallback(String res);
typedef void OnRegisterCallback(Map<String, dynamic> res);
typedef void OnErrorCallback(int code, String reason);
typedef void OnLocalStreamCallback(MediaStream stream);
typedef void OnRemoteStreamCallback(MediaStream stream, String refID);
typedef void OnReceiveCallFromUserCallback(String receivefrom, String type);
typedef void OnAudioVideoStateInfoCallback(
    int audioFlag, int videoFlag, String refID);
typedef void OnCallRejectedByUserCallback();
typedef void OnCallAcceptedByUserCallback();
typedef void OnCallHungUpByUserCallback();
typedef void OnCallDeclineByYouCallback();
typedef void OnCallBusyCallback();
typedef void OnParticipantsLeftCallback(String refID);

class SignalingClient {
  // IOWebSocketChannel _socket;
  static final SignalingClient _instance =
      SignalingClient._privateConstructor();
  static SignalingClient get instance => _instance;
  SignalingClient._privateConstructor();
  //callbacks
  OnConnectCallback onConnect;
  OnErrorCallback onError;
  OnRegisterCallback onRegister;
  OnLocalStreamCallback onLocalStream;
  OnRemoteStreamCallback onRemoteStream;
  OnReceiveCallFromUserCallback onReceiveCallFromUser;
  OnCallRejectedByUserCallback onCallRejectedByUser;
  OnCallAcceptedByUserCallback onCallAcceptedByUser;
  OnCallHungUpByUserCallback onCallHungUpByUser;
  OnCallDeclineByYouCallback onCallDeclineByYou;
  OnCallBusyCallback onCallBusyCallback;
  OnAudioVideoStateInfoCallback onAudioVideoStateInfo;
  OnParticipantsLeftCallback onParticipantsLeft;

  WebSocketConnection _socket;
  RTCPeerConnection _peerConnection;
  RTCPeerConnection _peerConnection2;
  Map<String, RTCPeerConnection> rtcPeerConnectionList = {};
  MediaStream _localStream;
  MediaStream _audioStream;
  MediaStream _audioInternalStream;
  // MediaStream stream;
  Map<String, RTCPeerConnection> requestIDType = {};
  String tenantID = "12345";
  String ref_id;
  String sessionUUID;
  bool screenShareflag = false;
  Stream sStream;
  String mcTokenCurrent;
  List<MediaStream> remoteStreams;
  static const MethodChannel _channel = const MethodChannel('flutter_vdotok');

  Future<dynamic> callAPI(auth_token, project_id) async {
    final datarequest = {"auth_token": auth_token, "project_id": project_id};

    var url = Uri.parse(host);
    final response = await http.post(
      url,
      body: json.encode(datarequest),
    );
    print("this is response of Api call ${json.decode(response.body)}");
    if (response.statusCode == 200) {
      print("${response.statusCode}");
      return json.decode(response.body);
    } else {
      throw Exception("Failed to Load Data");
    }
  }

  Stream<int> numberStream() async* {
    // for(int i = 1; i<= 10; i++) {
    //   await Future.delayed(Duration(seconds: 1));

    yield 20;
    // }
    yield await _channel.invokeMethod('getInternalAudio', {"audio": true});
  }

  Future<void> connect(auth_token, project_id) async {
    // Testing
    // _audioInternalStream   =
    //     await _channel.invokeMethod('getInternalAudio');
    // print("this is android native return ${_audioInternalStream.id}");
    // int errorCode = await MediaProjectionCreator.createMediaProjection();
    // if (errorCode == MediaProjectionCreator.ERROR_CODE_SUCCEED) {
    //   print("successed");
    // } else if (errorCode ==
    //     MediaProjectionCreator.ERROR_CODE_FAILED_USER_CANCELED) {
    //   print("Cancelled");
    // } else if (errorCode ==
    //     MediaProjectionCreator.ERROR_CODE_FAILED_SYSTEM_VERSION_TOO_LOW) {
    //   print("Failed: System API level need to higher than 21");
    // }
    var data = await callAPI(auth_token, project_id);
    print("this is data ${data}");
    _socket =
        WebSocketConnection(url: data['media_server_map']['complete_address']);

    _socket.onOpen = () {
      print('onOpen');
      onConnect.call("connected");
    };

    _socket.onMessage = (message) {
      // print('Received data: ' + message);
      onMessage(json.decode(message));
    };

    _socket.onClose = (int code, String reason) {
      print("this is onclose socket");
      onError.call(code, reason);
      // print('Closed by server [$code => $reason]!');
      // onSignalingStateChange?.call(SignalingState.ConnectionClosed);
    };

    await _socket.connect();
  }

  closeSocket() {
    if (_socket != null) _socket.close();
  }

  void onMessage(message) async {
    Map<String, dynamic> mapData = message;
    print("this is data id ${mapData["requestType"]}");
    print("this is data- ${mapData}");

    switch (mapData["requestType"]) {
      //for register
      case 'register':
        if (mapData["responseCode"] == 200) {
          // print("this is mcToken ${mapData["mcToken"]}");

          onRegister.call(mapData);
          mcTokenCurrent = mapData["mcToken"].toString();
        } else
          onError.call(mapData["responseCode"], mapData["responseMessage"]);
        break;
      case '401':
        onError.call(mapData["responseCode"], mapData["responseMessage"]);
        break;
      //   // case 'peers':
      //   //   {
      //   //     // List<dynamic> peers = data;
      //   //     // if (onPeersUpdate != null) {
      //   //     //   Map<String, dynamic> event = Map<String, dynamic>();
      //   //     //   event['self'] = _selfId;
      //   //     //   event['peers'] = peers;
      //   //     //   onPeersUpdate?.call(event);
      //   //     // }
      //   //   }
      //   //   break;
      case 'incomingCall':
        {
          sessionUUID = mapData["sessionUUID"];
          incomingCall(mapData);
        }
        break;
      case 'declineres':
        {
          onCallDeclineByYou.call();
          _peerConnection.close();
        }
        break;
      // startCommunication
      case 'new_participant_arrived':
        {
          handleNewParticipant(mapData);
        }
        break;
      case 'existing_participants':
        {
          handleExistingParticipant(mapData);
        }
        break;
      case 'startCommunication':
        {
          print(" this is start Communication ${rtcPeerConnectionList}");
          try {
            print("this is answer sdp ${mapData['sdpAnswer'].toString()}");
            if (mapData['referenceID'] == ref_id) {
              print("this is startCommunication with my ref_id");
              rtcPeerConnectionList[mapData['referenceID']]
                ..setRemoteDescription(
                    RTCSessionDescription(mapData['sdpAnswer'], "answer"));
            } else {
              print("this is startCommunication with other ref_id");
              rtcPeerConnectionList[mapData['referenceID']]
                ..setRemoteDescription(
                    RTCSessionDescription(mapData['sdpAnswer'], "answer"));
            }

            // RTCSessionDescription(answerSdp, "offer"));
          } catch (e) {
            print("this is error $e");
          }
        }
        break;
      case 'stopCommunication':
        {
          onCallHungUpByUser.call();
          _closeSession();
        }
        break;
      case 'session_invite':
        print("this is data after accepting, ${mapData["responseCode"]}");
        if (mapData["responseCode"] == 100) {
          print("trying to locate & validate target");
        } else if (mapData["responseCode"] == 180) {
          print("trying to connect target");
        } else if (mapData["responseCode"] == 183) {
          print("target is alerting");
        } else if (mapData["responseCode"] == 407) {
          print("authentication required");
        } else if (mapData["responseCode"] == 404) {
          print("invalid target");
          // onError.call(mapData["responseCode"], mapData["responseMessage"]);
          // _closeSession();
        } else if (mapData["responseCode"] == 480) {
          print("Temporarily Unavailable");
        } else if (mapData["responseCode"] == 403) {
          print("Service not allowed");
        } else if (mapData["responseCode"] == 408) {
          print("No Response from target");
          onError.call(mapData["responseCode"], mapData["responseMessage"]);
          _closeSession();
        } else if (mapData["responseCode"] == 486) {
          print("Target Busy");
          onCallBusyCallback.call();
          _closeSession();
        } else if (mapData["responseCode"] == 496) {
          print("rejected by user");
          onCallRejectedByUser.call();
          _closeSession();
        } else if (mapData["responseCode"] == 200) {
          print("accepted");
        }
        break;
      case 'callResponse':
        {
          try {
            rtcPeerConnectionList[ref_id]
              ..setRemoteDescription(
                  RTCSessionDescription(mapData['sdpAnswer'], "answer"));
            onCallAcceptedByUser.call();
          } catch (e) {
            print("this is error $e");
          }
        }
        break;
      case 'iceCandidate':
        {
          print(
              "this is data peer connection instance${rtcPeerConnectionList},   $mapData");
          var candidateMap = mapData['candidate'];
          print(
              "this is icecandidate ${candidateMap['candidate']} ${candidateMap['sdpMid']} ${candidateMap['sdpMLineIndex']}");

          //       String jsonString = sdpController.text;
          // dynamic session = await jsonDecode('$jsonString');
          // print(session['candidate']);
          try {
            dynamic candidate = new RTCIceCandidate(candidateMap['candidate'],
                candidateMap['sdpMid'], candidateMap['sdpMLineIndex']);
            print("this is icecandidate $candidate");

            if (mapData['referenceID'] == ref_id) {
              print(
                  "this is icecandidate with my ref_id ${rtcPeerConnectionList.containsKey(mapData['referenceID'])}    ${rtcPeerConnectionList[mapData['referenceID']]}");
              await rtcPeerConnectionList[mapData['referenceID']]
                  .addCandidate(candidate);
            } else {
              print(
                  "this is icecandidate with other ref_id ${rtcPeerConnectionList.containsKey(mapData['referenceID'])}    ${rtcPeerConnectionList[mapData['referenceID']]}");
              await rtcPeerConnectionList[mapData['referenceID']]
                  .addCandidate(candidate);
            }
          } catch (e) {
            print("this is error $e");
          }

          // RTCIceCandidate candidate = RTCIceCandidate(candidateMap['candidate'],
          //     candidateMap['sdpMid'], candidateMap['sdpMLineIndex']);

          // if (session != null) {
          //   if (session.pc != null) {
          //     await session.pc.addCandidate(candidate);
          //   } else {
          //     session.remoteCandidates.add(candidate);
          //   }
          // } else {
          //   _sessions[sessionId] = Session(pid: peerId, sid: sessionId)
          //     ..remoteCandidates.add(candidate);
          // }
        }
        break;

      case "session_cancel":
        {
          if (mapData["responseCode"] == 410) {
            _closeSessionLeftParticipant(mapData['referenceID']);
            onParticipantsLeft.call(mapData['referenceID']);
          } else {
            onCallHungUpByUser.call();
            _closeSession();
          }
        }
        break;

      case "state_information":
        {
          onAudioVideoStateInfo.call(mapData["audioInformation"],
              mapData["videoInformation"], mapData['referenceID']);
          // onCallHungUpByUser.call();
          // _closeSession();
        }
        break;
      //   // case 'leave':
      //   //   {
      //   //     var peerId = data as String;
      //   //     _closeSessionByPeerId(peerId);
      //   //   }
      //   //   break;
      //   // case 'bye':
      //   //   {
      //   //     var sessionId = data['session_id'];
      //   //     print('bye: ' + sessionId);
      //   //     var session = _sessions.remove(sessionId);
      //   //     onCallStateChange?.call(session, CallState.CallStateBye);
      //   //     _closeSession(session);
      //   //   }
      //   //   break;
      //   // case 'keepalive':
      //   //   {
      //   //     print('keepalive response!');
      //   //   }
      //   //   break;
      default:
        break;
    }
    if (mapData["id"] == "ping") {
      var response = {"requestType": "pong"};
      // const oneSec = const Duration(seconds: 2);
      // Timer.periodic(oneSec, (Timer t) => _socket.send(response));
      _socket.send(response);
    }
    //in case of busy user
    // if (mapData["responseCode"] == 486) {
    //   print("this is decline");
    //   onCallBusyCallback.call();
    //   _closeSession();
    // }
  }

  register(dynamic user, project_id) {
    ref_id = user["ref_id"];
    Map<String, dynamic> registerjson = {
      "type": "request",
      "requestType": "register",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "tenantID": project_id,
      "referenceID": user["ref_id"],
      "authorizationToken": user["authorization_token"],
      "socketType": 0
    };

    print("this is json $registerjson   $requestIDType");
    _socket.send(registerjson);
  }

  unRegister(String mc_token) {
    Map<String, dynamic> registerjson = {
      "type": "request",
      "requestType": "un_register",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "mcToken": mc_token
    };

    print("this is json $registerjson   $requestIDType");
    _socket.send(registerjson);
  }

  audioVideoState({int audioFlag, int videoFlag, String mcToken}) {
    print("this is statesssssss $audioFlag    $videoFlag");
    Map<String, dynamic> audioVideoStatejson = {
      "type": "request",
      "requestType": "state_information",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "referenceID": ref_id,
      "mcToken": mcToken,
      "sessionUUID": sessionUUID,
      "audioInformation": audioFlag,
      "videoInformation": videoFlag
    };
    print("this is state of audio and video json ${audioVideoStatejson}");

    _socket.send(audioVideoStatejson);
  }

  String generateMd5(String input) {
    return md5.convert(utf8.encode(input)).toString();
  }

  startCall(
      {String from,
      List<String> to,
      String mcToken,
      String meidaType,
      String callType,
      String sessionType}) async {
    screenShareflag = false;
    mcTokenCurrent = mcToken;
    try {
      print("this is sdk $mcToken");

      // RTCPeerConnection temp = await newParticipantsPeerConnection();

      // rtcPeerConnectionList.addAll({
      //   mapData["referenceID"]: {"peerConnection": temp},
      // });
      // _createOfferForNewParticipant(to: mapData["referenceID"].toString());
      // onReceiveCallFromUser.call(mapData["from"], mapData["media_type"]);
// Map<String, String> forSessionInvite = {
//   "requestType":"session_invite",
//   "media_type":meidaType,
//   "referenceID":from
// };

// RTCPeerConnection temp = await newParticipantsPeerConnection(forSessionInvite);

      await _createPeerConnection(type: meidaType);
      // rtcPeerConnectionList.addAll({
      //   ref_id: temp,
      // });
      _createOffer(
          from: from,
          to: to,
          mcToken: mcToken,
          mediaType: meidaType,
          callType: callType,
          sessionType: sessionType);
    } catch (e) {
      print("this is error $e");
    }
  }

  incomingCall(mapData) async {
    // RTCPeerConnection temp =
    await _createPeerConnection(type: mapData["media_type"]);
    // rtcPeerConnectionList.addAll({
    //   ref_id: temp,
    // });

    // RTCPeerConnection temp = await newParticipantsPeerConnection(mapData);

    // rtcPeerConnectionList.addAll({
    //   mapData["from"]: {"peerConnection": temp},
    // });
    // _createOfferForNewParticipant(to: mapData["from"].toString());
    onReceiveCallFromUser.call(mapData["from"], mapData["media_type"]);
  }

  handleExistingParticipant(mapData) async {
    print(
        "this is newparticipants data ${mapData["referenceIDs"].first},  $rtcPeerConnectionList");
    for (var i = 0; i < mapData["referenceIDs"].length; i++) {
      if (rtcPeerConnectionList.containsKey(mapData["referenceIDs"][i])) {
      } else {
        // OnRemoteStreamCallback remoteCallBack;
        // if (rtcPeerConnectionList.length == 1)
        //   remoteCallBack = onRemoteStream;
        // else {
        //   remoteCallBack = onRemoteStream1;
        // }
        await newParticipantsPeerConnection(
            mapData["referenceIDs"][i], onRemoteStream);
        _createOfferForNewParticipant(to: mapData["referenceIDs"][i]);
      }
    }
  }

  handleNewParticipant(mapData) async {
    print(
        "this is newparticipants data ${mapData["referenceID"]},  $rtcPeerConnectionList");
    // RTCPeerConnection temp =
    // OnRemoteStreamCallback remoteCallBack;
    // if (rtcPeerConnectionList.length == 1)
    //   remoteCallBack = onRemoteStream;
    // else {
    //   remoteCallBack = onRemoteStream1;
    // }

    await newParticipantsPeerConnection(mapData["referenceID"], onRemoteStream);

    // rtcPeerConnectionList.addAll({
    //   mapData["referenceID"]: temp,
    // });
    _createOfferForNewParticipant(to: mapData["referenceID"]);
  }

  newParticipantsPeerConnection(
      refID, OnRemoteStreamCallback remoteCallBack) async {
    try {
      Map<String, dynamic> configuration = {
        "iceServers": [
          {"url": "stun:stun.l.google.com:19302"},
        ]
      };

      final Map<String, dynamic> offerSdpConstraints = {
        "mandatory": {
          "OfferToReceiveAudio": true,
          "OfferToReceiveVideo": true,
        },
        "optional": [],
      };
      // if (mapData["requestType"] == "incomingCall") {
      // _localStream = await _getUserMedia(type: "video");
      // _localStream?.getVideoTracks()[0].enabled = false;
      print("this is local stream ");

      // onLocalStream.call(_localStream);
      // }

      rtcPeerConnectionList[refID] =
          await createPeerConnection(configuration, offerSdpConstraints);

      print(
          "this is create of second peer connection  $rtcPeerConnectionList ${rtcPeerConnectionList[refID]}");
      // var streams = pc.getLocalStreams();
      // for (var stream in streams) {
      //   _localStream = stream;
      // }
      // onLocalStream.call(_localStream);
      rtcPeerConnectionList[refID].addStream(_localStream);
      rtcPeerConnectionList[refID].onIceCandidate = (e) {
        if (e.candidate != null) {
          Map<String, dynamic> icCandidate = {
            "requestType": "onIceCandidate",
            "type": "request",
            "referenceID": refID,
            "sessionUUID": sessionUUID,
            "candidate": {
              "candidate": "${e.candidate}",
              "sdpMid": "${e.sdpMid}",
              "sdpMLineIndex": e.sdpMlineIndex.toString()
            }
          };

          print("these are on new participants icecandidates $icCandidate");
          _socket.send(icCandidate);
        }
      };

      rtcPeerConnectionList[refID].onIceConnectionState = (e) {};
      // for remote strem

      rtcPeerConnectionList[refID].onAddStream = (stream) {
        print(
            "this is remote stream ${stream.id}  ${rtcPeerConnectionList.length}");
        // remoteStreams.add(stream);

        // if (rtcPeerConnectionList.length == 2) {
        remoteCallBack.call(stream, refID);
        // } else if (rtcPeerConnectionList.length == 3) {
        //   onRemoteStream1.call(stream);
        // }
      };

      // return pc;
    } catch (e) {
      print("this is error $e");
    }
  }

  stopCall(mcToken, refID) {
    Map<String, dynamic> jsonData = {
      "type": "request",
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
    };
    _socket.send(jsonData);
    // _peerConnection.close();
    _closeSession();
  }

  _createOfferForNewParticipant({String to}) async {
    print("this is to $to  ${rtcPeerConnectionList[to]}");
    // print(
    //     "this is create offern lsdjflsjlfjsd ${rtcPeerConnectionList["${to.toString()}"]}");

    // RTCPeerConnection _temp = rtcPeerConnectionList["${to.toString()}"];
    try {
      RTCSessionDescription description =
          await rtcPeerConnectionList[to].createOffer({
        "OfferToReceiveAudio": 1,
        "OfferToReceiveVideo": 1,
      });
      await rtcPeerConnectionList[to].setLocalDescription(description);
      Map<String, dynamic> offerJson = {
        "from": ref_id,
        "to": to,
        "type": "request",
        "requestType": "to_receive_stream",
        "requestID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "sessionUUID": sessionUUID,
        "mcToken": mcTokenCurrent,
        "sdpOffer": description.sdp.toString()
      };
      // requestIDType[offerJson["requestID"]] = "call";
      print("this is json on create offer $offerJson");
      // sessionUUID = offerJson["sessionUUID"];
      // channel.sink.add(offerJson);
      _socket.send(offerJson);
    } catch (e) {
      print("this is error $e");
    }
  }

  _createOffer(
      {String from,
      List<String> to,
      String mcToken,
      String mediaType,
      String callType,
      String sessionType}) async {
    print("this is create offer $ref_id");
    try {
      RTCSessionDescription description =
          await rtcPeerConnectionList[ref_id].createOffer({
        "OfferToReceiveAudio": 1,
        "OfferToReceiveVideo": 1,
      });
      await rtcPeerConnectionList[ref_id].setLocalDescription(description);
      Map<String, dynamic> offerJson = {
        "from": from,
        "to": to,
        "type": "request",
        "requestType": "session_invite",
        "session_type": sessionType,
        "call_type": callType,
        "media_type": mediaType,
        "requestID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "sessionUUID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "mcToken": mcTokenCurrent,
        "sdpOffer": description.sdp.toString()
      };
      // requestIDType[offerJson["requestID"]] = "call";
      print("this is json on create offer $offerJson");
      sessionUUID = offerJson["sessionUUID"];
      // channel.sink.add(offerJson);
      _socket.send(offerJson);
    } catch (e) {
      print("this is error $e");
    }
  }

  createAnswer(refID) async {
    print("this is name $refID  ${rtcPeerConnectionList[ref_id]}");

    try {
      RTCSessionDescription description =
          await rtcPeerConnectionList[ref_id].createOffer({
        "OfferToReceiveAudio": 1,
        "OfferToReceiveVideo": 1,
      });
      // RTCSessionDescription description =
      //     await _peerConnection.createAnswer({'offerToReceiveVideo': 1});

      // var session = parse(description.sdp);
      // print(json.encode(session));
      // print(json.encode({
      //       'sdp': description.sdp.toString(),
      //       'type': description.type.toString(),
      //     }));

      rtcPeerConnectionList[ref_id].setLocalDescription(description);
      final Map<String, dynamic> _offerJson = {
        "type": "request",
        "requestID": generateMd5(
            DateTime.now().millisecondsSinceEpoch.toString() +
                tenantID +
                ref_id),
        "sessionUUID": sessionUUID,
        "requestType": "session_invite",
        "responseCode": 200,
        "responseMessage": "accepted",
        "sdpOffer": description.sdp.toString()
      };
      print("this is offer json $_offerJson");
      _socket.send(_offerJson);
      // channel.sink.add(_offerJson);
    } catch (e) {
      print("this is error $e");
    }
  }

  onDeclineCall(from, mcToken) {
    Map<String, dynamic> jsonData = {
      "type": "response",
      "from": from,
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
      // "responseCode": 487,
      // "responseMessage": "rejected"
    };

    print("this is json data $jsonData");
    _socket.send(jsonData);
    _closeSession();
  }

  onCancelbytheCaller(mcToken) {
    Map<String, dynamic> jsonData = {
      "type": "request",
      "requestType": "session_cancel",
      "requestID": generateMd5(
          DateTime.now().millisecondsSinceEpoch.toString() + tenantID + ref_id),
      "sessionUUID": sessionUUID,
      "mcToken": mcToken
    };
    _socket.send(jsonData);
    _closeSession();
  }

  _createPeerConnection({String type}) async {
    try {
      Map<String, dynamic> configuration = {
        "iceServers": [
          {"url": "stun:stun.l.google.com:19302"},
        ]
      };

      final Map<String, dynamic> offerSdpConstraints = {
        "mandatory": {
          "OfferToReceiveAudio": true,
          "OfferToReceiveVideo": true,
        },
        "optional": [],
      };

      _localStream = await _getUserMedia(type: type);
      // _localStream?.getVideoTracks()[0].enabled = false;
      print("this is local stream ");

      rtcPeerConnectionList[ref_id] =
          await createPeerConnection(configuration, offerSdpConstraints);

      print(
          "in create peer connection $rtcPeerConnectionList  ${rtcPeerConnectionList[ref_id]}");
      // var streams = pc.getLocalStreams();
      // for (var stream in streams) {
      //   _localStream = stream;
      // }
      onLocalStream.call(_localStream);
      rtcPeerConnectionList[ref_id].addStream(_localStream);
      rtcPeerConnectionList[ref_id].onIceCandidate = (e) {
        if (e.candidate != null) {
          Map<String, dynamic> icCandidate = {
            "requestType": "onIceCandidate",
            "type": "request",
            "referenceID": ref_id,
            "sessionUUID": sessionUUID,
            "candidate": {
              "candidate": "${e.candidate}",
              "sdpMid": "${e.sdpMid}",
              "sdpMLineIndex": e.sdpMlineIndex.toString()
            }
          };
          _socket.send(icCandidate);
          print(
              "these are on new participants icecandidates #!!!!!!!!!!!! $icCandidate");
        }
      };

      rtcPeerConnectionList[ref_id].onIceConnectionState = (e) {};
      // for remote strem

      // pc.onAddStream = (stream) {
      //   print("this is remote stream for first ${stream.id}");
      //   onRemoteStream.call(stream);
      // };

      // return _peerConnection;
    } catch (e) {
      print("this is error $e");
    }
  }

  void switchCamera() {
    if (_localStream != null) {
      _localStream.getVideoTracks()[0].switchCamera();
    }
  }

  void enableCamera(enabled) {
    print("this is bool $enabled");
    if (_localStream != null) {
      // _localStream.getVideoTracks()[0].switchCamera();
      _localStream?.getVideoTracks()[0].enabled = enabled;
    }
  }

  muteMic() {
    if (_localStream != null) {
      bool enabled = _localStream.getAudioTracks()[0].enabled;
      _localStream.getAudioTracks()[0].enabled = !enabled;
      print("this is local ${_localStream.getAudioTracks()[0].enabled}");
      return !enabled;
    }
  }

  switchSpeaker(flag) {
    if (_localStream != null) {
      _localStream?.getAudioTracks()[0].enableSpeakerphone(flag);
    }
  }

  _getUserMedia({String type}) async {
    try {
      Map<String, dynamic> mediaConstraints = {
        'audio': false,
        'video': {
          // "height": {"min": "812", "ideal": "1920"},
          'mandatory': {
            //  'minWidth': '640',
            //   "maxWidth": '640',
            //   'minHeight': '1319',
            //   "maxHeight": '640',

            //   "minWidth":'640', // Provide your own width, height and frame rate here
            // "minHeight": '480',
            // "minFrameRate": '30',
          },
          'facingMode': 'user',
          // 'optional': [],
        }
      };
      Map<String, dynamic> onlyAudio = {'audio': true};

      _audioStream =
          await navigator.mediaDevices.getUserMedia(onlyAudio).catchError((e) {
        print("this is error audio  $e");
      });
      if (type == "video") {
        return await
            // navigator.getDisplayMedia(mediaConstraints)
            navigator.mediaDevices.getUserMedia(mediaConstraints).then((value) {
          print("this is stream ${value}");
          _audioStream.getAudioTracks().forEach((element) {
            value.addTrack(element);
          });
          if (!kIsWeb) value?.getAudioTracks()[0].enableSpeakerphone(true);
          return value;
        }).catchError((onError) {
          print("this is error $onError");
        });
      } else {
        if (!kIsWeb)
          _audioStream?.getAudioTracks()[0].enableSpeakerphone(false);
        return _audioStream;
      }

      if (screenShareflag) {
        // return await navigator.getDisplayMedia(mediaConstraints)
        //     // navigator.mediaDevices.getUserMedia(mediaConstraints)
        //     .then((value) {
        //   print("this is stream ${value}");
        //   // _audioStream.getAudioTracks().forEach((element) {
        //   //   value.addTrack(element);
        //   // });
        //   return value;
        // }).catchError((onError) {
        //   print("this is error $onError");
        // });
      } else {
        // return await
        //     // navigator.getDisplayMedia(mediaConstraints)
        //     navigator.mediaDevices.getUserMedia(mediaConstraints).then((value) {
        //   print("this is stream ${value}");
        //   _audioStream.getAudioTracks().forEach((element) {
        //     value.addTrack(element);
        //   });
        //   return value;
        // }).catchError((onError) {
        //   print("this is error $onError");
        // });
        // if (Platform.isAndroid) {
        //   if (_localStream == null) {
        //     // define initial variable here for ability to call stop() on each track

        //     return await navigator.mediaDevices
        //         .getDisplayMedia(mediaConstraints)
        //         .then((displayStream) async {
        //       audioStream.getAudioTracks().forEach((element) {
        //         displayStream.addTrack(element);
        //       });
        //       return displayStream;
        //       // return await displayStream
        //       //     .addTrack(audioStream.getTracks()[0]); // do stuff
        //       // or
        //       // do stuff
        //     }).catchError((e) {
        //       print("this is error over all $e");
        //     });

        //     // return await navigator.mediaDevices
        //     //     .getUserMedia(mediaConstraints)
        //     //     .then((value) {
        //     //   print("this is stream ${value.id}");
        //     //   audioStream.getAudioTracks().forEach((element) {
        //     //     value.addTrack(element);
        //     //   });
        //     //   return value;
        //     // }).catchError((onError) {
        //     //   print("this is error $onError");
        //     // });
        //   } else {
        //     return _localStream;
        //   }
        // } else {
        //   print("teseting ");
        //   // ReplayKitLauncher.launchReplayKitBroadcast('sharescreen');
        //   return await
        //       // navigator.mediaDevices.getDisplayMedia(mediaConstraints);
        //       navigator.mediaDevices
        //           .getUserMedia(mediaConstraints)
        //           .then((value) {
        //     audioStream.getAudioTracks().forEach((element) {
        //       value.addTrack(element);
        //     });
        //     print("this is stream ${value.id}");
        //     return value;
        //   }).catchError((onError) {
        //     print("this is error $onError");
        //   });
        // }
      }

      // }
      // } else {
      //   stream = await navigator.mediaDevices
      //       .getUserMedia(mediaConstraints)
      //       .then((value) => value)
      //       .catchError((onError) {
      //     print("this is error $onError");
      //   });
      // }

      // _localStream = stream;
      // _localRenderer.srcObject = stream;
      // _localRenderer.mirror = true;

      // _peerConnection.addStream(stream);

      // return stream;
    } catch (e) {
      print("this is error $e");
    }
  }

  Future<void> _closeSession() async {
    print("this is close session");
    _localStream?.getTracks()?.forEach((element) async {
      await element.stop();
    });
    await _localStream?.dispose();
    _localStream = null;

    await _peerConnection?.close();
    await _peerConnection2?.close();

    rtcPeerConnectionList.forEach((key, value) {
      rtcPeerConnectionList[key]?.close();
    });
    rtcPeerConnectionList.clear();

    _audioStream?.getTracks()?.forEach((element) async {
      await element.stop();
    });
    await _audioStream?.dispose();
    _audioStream = null;

    screenShareflag = false;
    sessionUUID = null;
    requestIDType.clear();
  }

  Future<void> _closeSessionLeftParticipant(refID) async {
    print("this is close session");

    await rtcPeerConnectionList[refID]?.close();
    rtcPeerConnectionList.remove(refID);
  }
}
